class VizitigException(Exception): ...


class NoGraphException(VizitigException): ...


class GraphExists(VizitigException): ...


class UnknownIndexType(VizitigException): ...


class FormatInputError(VizitigException): ...


class VizIndexError(VizitigException): ...


class NoData(VizIndexError): ...


class IndexNotReady(VizIndexError): ...


class NoIndex(VizIndexError): ...


class KmerIndexNeeded(VizIndexError): ...


class IndexAlreadyExists(VizIndexError): ...


class InvalidIndexPath(VizIndexError): ...


class InvalidShardPath(VizIndexError): ...


class MissingShardPath(VizIndexError): ...


class IncompatibleViziGraph(VizitigException): ...


class NotAViziGraphError(VizitigException): ...


class FileExists(VizitigException): ...


class NameAlreadyTaken(VizitigException): ...


class CannotRetrieveGraphError(VizitigException): ...


class EmptyExportError(VizitigException): ...


class QueryError(VizitigException): ...


class ParseError(QueryError): ...


class KmerParsingError(ParseError): ...


class MetadataError(VizitigException): ...


class MetadataDuplicate(MetadataError): ...


class MetaNotFound(MetadataError): ...


class InvalidKmerSize(QueryError): ...


class GeneNotFound(QueryError): ...


class TranscriptNotFound(QueryError): ...


class MetaDataNotFound(QueryError): ...


class FixtureError(VizitigException): ...


class InvalidPath(VizitigException): ...


class AbundanceNotFound(VizitigException): ...
